//=====================================================================
//
//  File:      IBCPSession.cpp
//  Summary:   This sample reads a filestream column to a file using the 
//             IBCPSession interface and writes a format file.  
// 
//             Use the ICommandUpload or IRowsetFastLoadUpload Filestream samples 
//             to create the table that this sample reads from.
//             When this sample is run, you must pass a server name (or server\instance_name) 
//             and a name for the format file it will create.
// 
//  Date:      September 23, 2007
//
//---------------------------------------------------------------------
//
//  This file is part of the Microsoft SQL Server Code Samples.
//  Copyright (C) Microsoft Corporation.  All rights reserved.
//
//This source code is intended only as a supplement to Microsoft
//Development Tools and/or on-line documentation.  See these other
//materials for detailed information regarding Microsoft code samples.
//
//THIS CODE AND INFORMATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF 
//ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO 
//THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
//PARTICULAR PURPOSE.
//
//===================================================================== 

#include "stdafx.h"

#define SAFE_RELEASE(x)    { if (x) { x->Release(); x = NULL; } }
#define SAFE_SYSFREE(bstr) { if (bstr) { ::SysFreeString(bstr); bstr = NULL; } }
#define CHECK_HR(hr,jump)  { if ( FAILED(hr) ) goto jump; }
#define CHECK_HR_ERR(hr,jump)  { if ( FAILED(hr) ) { DumpErrorRecords(); goto jump; } }

// Helper function for DumpErrorRecords() to convert a clisd to string representation of clsid.
LPWSTR CLSIDToString( CLSID* pclsid ) {
   static WCHAR wszCLSID[1024];
   if ( NULL == pclsid )
      wcscpy_s(wszCLSID, 1024, L"(Null)");   
   else
      ::StringFromGUID2( *pclsid, (OLECHAR*)wszCLSID, sizeof(wszCLSID) );
   return wszCLSID;
}

// Helper function to dump out OLE DB errors to console window.
HRESULT DumpErrorRecords() {
   HRESULT hr = S_OK;
   IErrorRecords* pIErrorRecords = NULL;
   IErrorInfo* pIErrorInfo = NULL;
   IErrorInfo* pIErrorInfoRecord = NULL;
   BSTR bstrDescription = NULL;
   BSTR bstrSource = NULL;
   ULONG i;
   ULONG cRecords = 0;
   LCID lcid = 0;
   ERRORINFO ErrorInfo;

   // Get IErrorInfo pointer from OLE.
   hr = ::GetErrorInfo( 0, &pIErrorInfo );
   if ( FAILED(hr) ) {
      printf( "No error records found.\n" );
      goto DumpErrorRecordsExit;
   }

   // QI for IID_IErrorRecords.
   hr = pIErrorInfo->QueryInterface( IID_IErrorRecords, (void**)&pIErrorRecords );
   if ( FAILED(hr) ) {
      printf( "pIErrorInfo->QueryInterface(IID_IErrorRecords) failed with hr=0x%08x.\n", hr );
      goto DumpErrorRecordsExit;
   }

   // Get error record count.
   hr = pIErrorRecords->GetRecordCount( &cRecords );
   if ( FAILED(hr) ) {
      printf( "pIErrorRecords->GetRecordCount() failed with hr=0x%08x.\n", hr );
      goto DumpErrorRecordsExit;
   }

   // Get system LCID
   lcid = GetSystemDefaultLCID(); 

   // Loop through the error records.
   for ( i = 0 ; i < cRecords ; i++ ) {
      // Get pIErrorInfo from pIErrorRecords.
      hr = pIErrorRecords->GetErrorInfo( i, lcid, &pIErrorInfoRecord );

      if ( SUCCEEDED(hr) ) {
         // Get error description and source.
         hr = pIErrorInfoRecord->GetDescription( &bstrDescription );
         hr = pIErrorInfoRecord->GetSource( &bstrSource );

         // Retrieve the ErrorInfo structures.
         hr = pIErrorRecords->GetBasicErrorInfo( i, &ErrorInfo );

         // Dump out error.
         printf( "\n" );
         printf( "Dumping error record %lu of %lu:\n", i+1, cRecords );
         printf( "--------------------------------------------------------------------------\n" );
         printf( "   Description = %S\n", bstrDescription );
         printf( "   Source      = %S\n", bstrSource );
         printf( "   ERRORINFO.hrError = 0x%08x\n", ErrorInfo.hrError );
         printf( "   ERRORINFO.dwMinor = %lu\n",	ErrorInfo.dwMinor ); 
         printf( "   ERRORINFO.clsid   = %S\n",	CLSIDToString( &ErrorInfo.clsid ) );

         SAFE_RELEASE(pIErrorInfoRecord);
         SAFE_SYSFREE(bstrDescription);
         SAFE_SYSFREE(bstrSource);
      }
   }

DumpErrorRecordsExit:
   SAFE_RELEASE(pIErrorInfoRecord);
   SAFE_RELEASE(pIErrorRecords);
   SAFE_RELEASE(pIErrorInfo);
   SAFE_RELEASE(pIErrorInfo);
   SAFE_SYSFREE(bstrDescription);
   SAFE_SYSFREE(bstrSource);
   return hr;
}

// to open a session

// 1) create a properties object
// 2) set properties on it (such as INIT_DATASOURCE, etc.)
// 3) queryinterface an IDBInitialize on the properties object
// 4) call IDBInitialize::Initialize
// 5) queryinterface an IDBCreateSession on the IDBInitialize instance
// 6) call IDBCreateSession::CreateSession with interface to use on the session (IDBCreateCommand, IOpenRowset, IBCPSession, etc.)

HRESULT OpenSession( wchar_t* datasource, wchar_t* catalog, const IID& iid, void** out ) {
#define SET_BSTR_PROP(index, proparray, propid, value) \
   { proparray[index].dwOptions = DBPROPOPTIONS_REQUIRED; \
   proparray[index].colid = DB_NULLID; \
   proparray[index].dwStatus = DBPROPSTATUS_OK; \
   proparray[index].dwPropertyID = propid; \
   proparray[index].vValue.vt = VT_BSTR; \
   proparray[index].vValue.bstrVal = SysAllocString(value); }

#define SET_BOOL_PROP(index, proparray, propid, value) \
   { proparray[index].dwOptions = DBPROPOPTIONS_REQUIRED; \
   proparray[index].colid = DB_NULLID; \
   proparray[index].dwStatus = DBPROPSTATUS_OK; \
   proparray[index].dwPropertyID = propid; \
   proparray[index].vValue.vt = VT_BOOL; \
   proparray[index].vValue.boolVal = (value) ? VARIANT_TRUE : VARIANT_FALSE; }

   HRESULT hr;

   IDBProperties*     pIDBProperties = NULL;
   IDBProperties*     pIDBSessionProperties = NULL;
   IDBInitialize*     pIDBInitialize = NULL;  
   IDBCreateSession*  pIDBCreateSession = NULL;

   DBPROP      InitProperties[3];
   DBPROP      BulkCopyProperties[2];
   DBPROPSET   PropSet[2];

   hr = CoCreateInstance( SQLNCLI_CLSID, NULL, CLSCTX_INPROC_SERVER, IID_IDBProperties, 
      (void **) &pIDBProperties );    
   CHECK_HR(hr,OpenSessionCleanup);

   // Hard coded this to point to local sql server, change as needed.
   SET_BSTR_PROP( 0, InitProperties, DBPROP_INIT_DATASOURCE, datasource );
   SET_BSTR_PROP( 1, InitProperties, DBPROP_INIT_CATALOG, catalog );
   SET_BSTR_PROP( 2, InitProperties, DBPROP_AUTH_INTEGRATED, L"SSPI" );

   SET_BOOL_PROP( 0, BulkCopyProperties, SSPROP_ENABLEFASTLOAD, true );
   SET_BOOL_PROP( 1, BulkCopyProperties, SSPROP_ENABLEBULKCOPY, true );

   PropSet[0].guidPropertySet = DBPROPSET_DBINIT; 
   PropSet[0].cProperties = 3;               
   PropSet[0].rgProperties = InitProperties; 

   PropSet[1].guidPropertySet = DBPROPSET_SQLSERVERDATASOURCE;
   PropSet[1].cProperties = 2;
   PropSet[1].rgProperties = BulkCopyProperties;

   pIDBProperties->SetProperties( 1, (DBPROPSET*) &PropSet );
   CHECK_HR( hr, OpenSessionCleanup );

   hr = pIDBProperties->QueryInterface( IID_IDBInitialize, (void**) &pIDBInitialize );
   CHECK_HR( hr,OpenSessionCleanup );

   hr = pIDBInitialize->Initialize();
   CHECK_HR( hr,OpenSessionCleanup );

   // QI for IID_IDBCreateSession and then create session.
   hr = pIDBInitialize->QueryInterface( IID_IDBCreateSession, (void**) &pIDBCreateSession );
   CHECK_HR( hr,OpenSessionCleanup );

   pIDBCreateSession->QueryInterface( IID_IDBProperties, (void**) &pIDBSessionProperties );
   pIDBSessionProperties->SetProperties( 1, &PropSet[1] );

   hr = pIDBCreateSession->CreateSession( NULL, iid, (IUnknown**) out );

OpenSessionCleanup:
   SAFE_RELEASE( pIDBProperties );
   SAFE_RELEASE( pIDBInitialize );
   SAFE_RELEASE( pIDBCreateSession );
   SAFE_RELEASE( pIDBSessionProperties );
   return hr;    	
}

int wmain(int argc, wchar_t* argv[]) {
   HRESULT       hr;
   IBCPSession*  pIBCPSession = NULL;
   DBROWCOUNT    rows;

   CoInitialize(NULL);

   if( argc < 2 ) {
      printf( "Usage: IBCPSession <server> <file> [<error file>]\n");
      return 0;
   }

   // Open connection to database.
   hr = OpenSession( argv[1], L"master", IID_IBCPSession, (void**) &pIBCPSession );
   CHECK_HR_ERR(hr,MainCleanup);

   // change the last parameter to BCP_DIRECTION_IN to make it write a file into the server
   hr = pIBCPSession->BCPInit( L"[DBFsa]..[FSTRTable]", argv[2], (argc > 3) ? argv[3] : NULL, BCP_DIRECTION_OUT );
   CHECK_HR_ERR( hr, MainCleanup );

#define WRITE_FMT_FILE 1
#if WRITE_FMT_FILE

   hr = pIBCPSession->BCPColumns( 3 );
   CHECK_HR_ERR( hr, MainCleanup );

   hr = pIBCPSession->BCPColFmt( 1, BCP_TYPE_DEFAULT, BCP_PREFIX_DEFAULT, BCP_LENGTH_VARIABLE, NULL, 0, 1 );
   CHECK_HR_ERR( hr, MainCleanup );

   hr = pIBCPSession->BCPColFmt( 2, BCP_TYPE_DEFAULT, BCP_PREFIX_DEFAULT, BCP_LENGTH_VARIABLE, NULL, 0, 2 );
   CHECK_HR_ERR( hr, MainCleanup );

   hr = pIBCPSession->BCPColFmt( 3, BCP_TYPE_DEFAULT, BCP_PREFIX_DEFAULT, BCP_LENGTH_VARIABLE, NULL, 0, 3 );
   CHECK_HR_ERR( hr, MainCleanup );

   hr = pIBCPSession->BCPWriteFmt( L"fstrtable.fmt" );
   CHECK_HR_ERR( hr, MainCleanup );

#else

   hr = pIBCPSession->BCPReadFmt( L"fstrtable.fmt" );
   CHECK_HR_ERR( hr, MainCleanup );

#endif

   hr = pIBCPSession->BCPExec( &rows );
   CHECK_HR_ERR( hr, MainCleanup );

   hr = pIBCPSession->BCPDone();
   CHECK_HR_ERR( hr, MainCleanup );

MainCleanup:
   SAFE_RELEASE( pIBCPSession );

   printf_s( "Test complete.\n" );
}